/*

Tkwav

Copyright (c) 1998, Nat Pryce

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in
all copies of this software.

IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, 
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF 
THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" 
BASIS, AND THE AUTHOR HAS NO OBLIGATION TO PROVIDE  MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

*/

#include <memory.h>
#include <functional>
#include <sstream>

#include "audio.h"


using namespace std;


/*-----------------------------------------------------------------------------
 *  class: audio_error
 */

void audio_error::raise( HRESULT status, std::string msg ) 
	throw(audio_error) 
{
	switch( status )
	{
	case DSERR_ALLOCATED :
		msg += ": the request failed because resources, such as a priority level, were already in use by another caller";
		break;
	case DSERR_ALREADYINITIALIZED :
		msg += ": the object is already initialized";
		break;
	case DSERR_BADFORMAT :
		msg += ": the specified wave format is not supported";
		break;
	case DSERR_BUFFERLOST :
		msg += ": the buffer memory has been lost and must be restored";
		break;
	case DSERR_CONTROLUNAVAIL :
		msg += ": the control (volume, pan, and so forth) requested by the caller is not available";
		break;
	case DSERR_INVALIDCALL :
		msg += ": this function is not valid for the current state of this object";
		break;
	case DSERR_NODRIVER :
		msg += ": no sound driver is available for use";
		break;
	case DSERR_OTHERAPPHASPRIO :
		msg += ": another application has a higher priority level, preventing this call from succeeding";
		break;
	case DSERR_PRIOLEVELNEEDED :
		msg += ": the caller does not have the priority level required for the function to succeed";
		break;
	case DSERR_UNINITIALIZED :
		msg += ": the audio device has not been initialised";
		break;

	default: 
		{
			ostringstream str( msg, ios_base::app );
			str << " (hresult = " << status << ")";
		}
	}
	
	throw audio_error( status, msg );
}


/*-----------------------------------------------------------------------------
 *  class: OutputDevice
 */

OutputDevice::OutputDevice() 
	throw(audio_error) 
{
	audio_error::check( DirectSoundCreate( NULL, &_device, NULL ),
						"failed to initialise sound driver" );
}

OutputDevice::OutputDevice( HWND window, CooperativeLevel level ) 
	throw(audio_error) 
{
	audio_error::check( DirectSoundCreate( NULL, &_device, NULL ),
						"failed to initialise sound driver" );
	cooperate(window,level);
}

void OutputDevice::cooperate( HWND window, CooperativeLevel level ) {
	HRESULT status = _device->SetCooperativeLevel( 
						 window, 
						 static_cast<DWORD>(level) );

	audio_error::check( status, "failed to set cooperative level" );
}

OutputDevice::~OutputDevice() {
	_device->Release();
}




/*-----------------------------------------------------------------------------
 *  class: SoundDescriptor
 */

SoundDescriptor::SoundDescriptor( DWORD buffer_size ) {
	memset( this, 0, sizeof(*this) );
	dwSize = sizeof(DSBUFFERDESC);
	dwFlags = DSBCAPS_CTRLDEFAULT|DSBCAPS_STATIC;
	dwBufferBytes = buffer_size;
	lpwfxFormat = &_format;
	
	_format.cbSize = sizeof(_format);
	_format.wFormatTag = WAVE_FORMAT_PCM;
	_format.nChannels = 1;
	_format.nSamplesPerSec = 22050;
	_format.wBitsPerSample = 8;
	
	updateFields();
}

SoundDescriptor::SoundDescriptor( const PCMWAVEFORMAT &fmt, DWORD buffer_size ) {
	memset( this, 0, sizeof(*this) );
	dwSize = sizeof(DSBUFFERDESC);
	dwFlags = DSBCAPS_CTRLDEFAULT|DSBCAPS_STATIC;
	dwBufferBytes = buffer_size;
	lpwfxFormat = &_format;
	
	memcpy( &_format, &fmt, sizeof(_format) );
	
	updateFields();
}


bool SoundDescriptor::controlFrequency() const {
	return dwFlags & DSBCAPS_CTRLFREQUENCY != 0;
}

bool SoundDescriptor::controlPan() const {
	return dwFlags & DSBCAPS_CTRLPAN != 0;
}

bool SoundDescriptor::controlVolume() const {
	return dwFlags & DSBCAPS_CTRLVOLUME != 0;
}

bool SoundDescriptor::isStatic() const {
	return dwFlags & DSBCAPS_STATIC != 0;
}

unsigned SoundDescriptor::channels() const {
	return _format.nChannels;
}

unsigned long SoundDescriptor::rate() const {
	return _format.nSamplesPerSec;
}

unsigned long SoundDescriptor::sampleSize() const {
	return _format.wBitsPerSample;
}

unsigned long SoundDescriptor::bufferSize() const {
	return dwBufferBytes;
}

SoundDescriptor &SoundDescriptor::controlFrequency( bool f ) {
	if(f) {
		dwFlags |= DSBCAPS_CTRLFREQUENCY;
	} else {
		dwFlags &= ~DSBCAPS_CTRLFREQUENCY;
	}
	return *this;
}

SoundDescriptor &SoundDescriptor::controlPan( bool f ) {
	if(f) {
		dwFlags |= DSBCAPS_CTRLPAN;
	} else {
		dwFlags &= ~DSBCAPS_CTRLPAN;
	}
	return *this;
}

SoundDescriptor &SoundDescriptor::controlVolume( bool f ) {
	if(f) {
		dwFlags |= DSBCAPS_CTRLVOLUME;
	} else {
		dwFlags &= ~DSBCAPS_CTRLVOLUME;
	}
	return *this;
}

SoundDescriptor &SoundDescriptor::isStatic( bool f ) {
	if(f) {
		dwFlags |= DSBCAPS_STATIC;
	} else {
		dwFlags &= ~DSBCAPS_STATIC;
	}
	return *this;
}

SoundDescriptor &SoundDescriptor::channels( unsigned n ) {
	_format.nChannels = n;
	updateFields();
	return *this;
}

SoundDescriptor &SoundDescriptor::rate( unsigned long hz ) {
	_format.nSamplesPerSec = hz;
	updateFields();
	return *this;
}

SoundDescriptor &SoundDescriptor::sampleSize( unsigned long size ) {
	_format.wBitsPerSample = size;
	updateFields();
	return *this;
}

SoundDescriptor &SoundDescriptor::bufferSize( unsigned long size ) {
	dwBufferBytes = size;
	return *this;
}

void SoundDescriptor::updateFields() {
	_format.nBlockAlign = _format.nChannels * _format.wBitsPerSample/8;
	_format.nAvgBytesPerSec = _format.nBlockAlign * _format.nSamplesPerSec;
}



/*-----------------------------------------------------------------------------
 *  class: Sound
 */

Sound::Sound( OutputDevice &dev, const SoundDescriptor &desc,
			  void *data, size_t size ) throw(audio_error) :
	_device(dev),
	_buffer(0)
{
	HRESULT status;
	void *buffer;
	DWORD size1;
	DWORD size2;	// always set to 0 by _buffer->Lock

	
	status = dev._device->CreateSoundBuffer( 
		const_cast<SoundDescriptor*>(&desc), 
		&_buffer, 
		NULL );
	audio_error::check( status, "failed to create sound buffer" );

	status = _buffer->Lock( 0,				// Lock from start of buffer
							this->size(),	// Lock entire buffer
							&buffer,		// Sets data to point to sound data
							&size1,			// Sets size to the amount of data
							NULL,			// ignored
							&size2,			// ignored
							0 );			// No flags
	audio_error::check( status, "failed to lock sound buffer" );
	
	memcpy( buffer, data, min(size,size1) );
	
	audio_error::check( _buffer->Unlock( buffer, size1, NULL, 0 ),
						"failed to unlock sound buffer" );
}

Sound::Sound( const Sound &s ) throw(audio_error) :
	_device(s._device),
	_buffer(0)
{
	HRESULT status;

	status = _device._device->DuplicateSoundBuffer( s._buffer, &_buffer );
	audio_error::check( status, "failed to duplicate sound buffer" );
}

bool Sound::isPlaying() const throw(audio_error) {
	DWORD state;
	HRESULT status = _buffer->GetStatus(&state);

	audio_error::check( status, "failed to get play state" );

	return state == DSBSTATUS_PLAYING || state == DSBSTATUS_LOOPING;
}

unsigned long Sound::size() const throw(audio_error) {
	DSBCAPS caps;

	caps.dwSize = sizeof(caps);
	audio_error::check( _buffer->GetCaps(&caps),
						"failed to get buffer capabilities" );

	return static_cast<unsigned long>(caps.dwBufferBytes);
}

void Sound::play() throw(audio_error) {
	audio_error::check( _buffer->Play( 0, 0, 0 ),
						"failed to play sound" );
}

void Sound::loop() throw(audio_error) {
	audio_error::check( _buffer->Play( 0, 0, DSBPLAY_LOOPING ),
						"failed to loop sound" );
}

void Sound::pause() throw(audio_error) {
	audio_error::check( _buffer->Stop(),
						"failed to pause sound" );
}

void Sound::stop() throw(audio_error) {
	pause();
	audio_error::check( _buffer->SetCurrentPosition(0),
						"failed to rewind sound" );
}

Sound::~Sound() {
	_buffer->Stop();
	_buffer->Release();
}

