/*

Tkwav

Copyright (c) 1998, Nat Pryce

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in
all copies of this software.

IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, 
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF 
THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" 
BASIS, AND THE AUTHOR HAS NO OBLIGATION TO PROVIDE  MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

*/
#ifndef _AUDIO_H
#define _AUDIO_H


#include <stdexcept>
#include <windows.h>
#include <dsound.h>


/*  Base class of exceptions that can be thrown by Sound objects.
 */
class audio_error : public std::runtime_error
{
public:
    audio_error( HRESULT error, const std::string &msg ) :
        std::runtime_error(msg), _error(error) {
    }
	
    HRESULT status() const throw() { return _error; }
    unsigned int code() const throw() { return HRESULT_CODE(_error); }
    unsigned int facility() const throw() { return HRESULT_FACILITY(_error); }
    unsigned int severity() const throw() { return HRESULT_SEVERITY(_error); }
	
	/*  Efficiently checks whether 'status' indicates an error and
	 *  throws an audio_error exception if it does.
	 */
    static inline
    void check( HRESULT status, const char *msg="" ) throw(audio_error) {
        if( FAILED(status) ) raise(status,msg);
    }
	
	/*  Throws an audio_error with an explanatory error message
	 *  that is made up of 'msg' followed by a description of
	 *  the 'status' code.
	 */
	static
	void raise( HRESULT status, std::string msg ) throw(audio_error);
	
private:
    HRESULT _error;
};



/*  An interface to an audio device
 */
class OutputDevice
{
public:
	enum CooperativeLevel {
		NORMAL			= DSSCL_NORMAL,	
		EXCLUSIVE		= DSSCL_EXCLUSIVE, 
		PRIORITY		= DSSCL_PRIORITY, 
		WRITE_PRIMARY	= DSSCL_WRITEPRIMARY
	};

	OutputDevice() throw(audio_error);
	OutputDevice( HWND win, CooperativeLevel = NORMAL ) 
		throw(audio_error);
	
	void cooperate( HWND win, CooperativeLevel = NORMAL ) 
		throw(audio_error);
	
	~OutputDevice();

private:
	friend class Sound;
	IDirectSound *_device;

};


/*  Dumb class that describes the required properties of a new Sound
 */
class SoundDescriptor : public DSBUFFERDESC
{
public:
	SoundDescriptor( const PCMWAVEFORMAT &fmt, unsigned long buffer_size );
	SoundDescriptor( unsigned long buffer_size );
	
	/*  Accessors
	 */
	bool controlFrequency() const;		// Default = true
	bool controlPan() const;			// Default = true
	bool controlVolume() const;			// Default = true
	bool isStatic() const;				// Default = true
	unsigned channels() const;			// Default = 1 or from fmt
	unsigned long rate() const;			// Default = 22050 Hz or from fmt
	unsigned long sampleSize() const;	// Default = 8 or from fmt
	unsigned long bufferSize() const;	// Set by constructor
	
	/*  Modifiers
	 */
	SoundDescriptor &controlFrequency( bool );
	SoundDescriptor &controlPan( bool );
	SoundDescriptor &controlVolume( bool );
	SoundDescriptor &isStatic( bool );
	SoundDescriptor &channels( unsigned n );
	SoundDescriptor &rate( unsigned long hz );
	SoundDescriptor &sampleSize( unsigned long size );
	SoundDescriptor &bufferSize( unsigned long size );
	
	
private:
	WAVEFORMATEX _format;

	void updateFields(); // Calculate format fields that rely on others
};



/*  A sound sample that can be played, paused, looped etc.
 */
class Sound
{
public:
	Sound( OutputDevice &device, const SoundDescriptor &desc, 
		   void *data, size_t size ) throw(audio_error);
	Sound( const Sound &sound ) 
		throw(audio_error);
	
	/*  Accessors
	 */
	bool isPlaying() const throw(audio_error);
	unsigned long size() const;
	
	/*  Control playing of sound
	 */
	void play() throw(audio_error);		// Start or resume playing
	void loop() throw(audio_error);		// Start looping play
	void pause() throw(audio_error);	// Pause play
	void stop() throw(audio_error);		// Stop play and rewind to start

	~Sound();
	
private:
	OutputDevice &_device;
	IDirectSoundBuffer *_buffer;
};


#endif // _AUDIO_H

