/*

Tkwav

Copyright (c) 1998, Nat Pryce

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in
all copies of this software.

IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, 
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF 
THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" 
BASIS, AND THE AUTHOR HAS NO OBLIGATION TO PROVIDE  MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

*/
#ifndef _AUTOPTR_H
#define _AUTOPTR_H


#include <functional>

/*  AutoPtr<T> : a smart pointer which automatically destroys what it is
 *               pointing at when it is destroyed itself.  The method of
 *               destruction can be specified as a template parameter,
 *               but defaults to deleting the object.
 */

template <class T>
struct Delete : public std::unary_function<T*,void> {
    void operator() ( T *ptr ) { delete ptr; }
};


template <typename T, typename DEL_FN = Delete<T> >
class AutoPtr
{
public:
    AutoPtr( const DEL_FN &destroy = DEL_FN() ) : 
        _ptr(0), _destroy(destroy) {
    }
    AutoPtr( T *t, const DEL_FN &destroy = DEL_FN() ) : 
        _ptr(t), _destroy(destroy) {
    }

    const AutoPtr<T> &operator = ( T *t ) {
        T *temp = _ptr;
        _ptr = t;
        if(temp) _destroy(temp);
        return *this;
    }

    operator bool () const { return _ptr != 0; }
    T &operator*  () const { return *_ptr; }
    T *operator-> () const { return _ptr; }

    ~AutoPtr() { if(_ptr) _destroy(_ptr); }

private:
    T *_ptr;
    DEL_FN _destroy;
};

#endif _AUTOPTR_H 