/*

Tkwav

Copyright (c) 1998, Nat Pryce

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in
all copies of this software.

IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, 
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF 
THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" 
BASIS, AND THE AUTHOR HAS NO OBLIGATION TO PROVIDE  MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

*/


#include <stdio.h>
#include <windows.h>

#include <tk.h>
#include "tkWin.h" //  This should be in the standard Tcl/Tk include directory
#include <hash.h>

#include "tkwav.h"
#include "audio.h"
#include "wavio.h"
#include "autoptr.h"



/*  Structures holding information about a loaded sound.
 */

struct FileInfo {
	FileInfo( Sound* s ) : sound(s), shared(0) {}

	AutoPtr<Sound> sound;
	unsigned int shared;
};

struct SoundInfo {
	SoundInfo( FileInfo &fi, string n, string f ) : 
		sound(new Sound(*fi.sound)), 
		name(n), 
		file(f) 
	{
		fi.shared++;
	}
	
	AutoPtr<Sound> sound;
	string name;
	string file;
};


/*  Static tables indexing loaded sounds by name or file
 */
static AutoPtr<OutputDevice> device;
static Tk_Window coop_window = NULL;
static OutputDevice::CooperativeLevel coop_level = OutputDevice::NORMAL;
static Hashtable<SoundInfo> sounds_by_name;
static Hashtable<FileInfo> sounds_by_file;


/*
 * Declarations for functions defined in this file.
 */

static int LoadCmd( ClientData clientData, Tcl_Interp *interp,
                    int objc, Tcl_Obj *CONST objv[] );
static int SoundCmd( ClientData clientData, Tcl_Interp *interp,
                     int objc, Tcl_Obj *CONST objv[] );
static int DeleteCmd( ClientData clientData, Tcl_Interp *interp,
                      int objc, Tcl_Obj *CONST objv[] );
static int CoopCmd( ClientData clientData, Tcl_Interp *interp,
                    int objc, Tcl_Obj *CONST objv[] );

static int ReportException( Tcl_Interp *interp, std::exception &e );
static void DeleteSound( ClientData );
static HWND ToplevelWindow( Tk_Window );

/*
 *----------------------------------------------------------------------
 *
 * DllEntryPoint --
 *
 *	This wrapper function is used by Windows to invoke the
 *	initialization code for the DLL.  If we are compiling
 *	with Visual C++, this routine will be renamed to DllMain.
 *	routine.
 *
 * Results:
 *	Returns TRUE;
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

extern "C" 
BOOL APIENTRY DllEntryPoint( HINSTANCE, DWORD, LPVOID ) {
	return TRUE;
}

/*
 *----------------------------------------------------------------------
 *
 * Tkwav_Init --
 *
 *	This procedure initializes the uuid command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

extern "C" 
EXPORT(int,Tkwav_Init)( Tcl_Interp *interp ) {
	coop_window = Tk_MainWindow(interp);
	Tk_MakeWindowExist(coop_window);
	
	/*  Initialise device to use the main Tk window for cooperation.
	 */
	if( !device ) try {
		device = new OutputDevice( ToplevelWindow(coop_window), coop_level );
	}
	catch( audio_error &e ) {
		return ReportException( interp, e );
	}
	
	/*  Create commands
	 */
	Tcl_CreateObjCommand( interp, "tkwav::load", LoadCmd, NULL, NULL );
	Tcl_CreateObjCommand( interp, "tkwav::destroy", DeleteCmd, NULL, NULL );
	Tcl_CreateObjCommand( interp, "tkwav::cooperate", CoopCmd, NULL, NULL );
	
	/*  Export commands from the tkwav namespace
	 */
	Tcl_GlobalEval( interp, 
		"namespace eval tkwav { namespace export load delete cooperate }" );
	
    return Tcl_PkgProvide(interp, "tkwav", "1.0");
}


/*
 *----------------------------------------------------------------------
 *
 * Tkwav_SafeInit --
 *
 *	This procedure initializes the example command for a safe 
 *	interpreter.  You would leave out of this procedure any commands
 *	you deemed unsafe.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

extern "C" 
EXPORT(int,Tkwav_SafeInit)( Tcl_Interp *interp ) {
	return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * LoadCmd --
 *
 *	This function implements the "tkwav::load" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Sets the result of interp to hold the name of the new command used
 *	to control the sound object.
 *	Adds the new sound to the the sounds-by-name and sounds-by-file
 *	tables
 *
 *----------------------------------------------------------------------
 */
static 
int LoadCmd( ClientData clientData, Tcl_Interp *interp, 
             int objc, Tcl_Obj *CONST objv[] )
{	
	char *name = 0;
	int name_len = 0;
	char *file = 0;
	int file_len;
	SoundInfo *info = 0;
	
	if( objc != 3 ) {
		Tcl_WrongNumArgs( interp, 1, objv, " name pathname" );
		return TCL_ERROR;
	}
	
	name = Tcl_GetStringFromObj( objv[1], &name_len );
	file = Tcl_GetStringFromObj( objv[2], &file_len );
		
	
	/*  Check for name-clash
	 */
	if( sounds_by_name.find(name) != NULL ) {
		Tcl_Obj *msg = Tcl_NewStringObj( "a sound named \"", -1 );
		Tcl_AppendToObj( msg, name, name_len );
		Tcl_AppendToObj( msg, "\" already exists", -1 );
		Tcl_SetObjResult( interp, msg );
		return TCL_ERROR;
	}
	

	/*  Check if the file has already been loaded.  If it has,
	 *  duplicate a sound that was previously loaded, otherwise
	 *  cache the sound in case the file is loaded again. This 
	 *  makes efficient use of the memory on the sound card.
	 */
	FileInfo *cache = sounds_by_file.find(file);
	if( cache == NULL ) {
		cache = new FileInfo( LoadWavFile(*device,file) );
		sounds_by_file.put( file, cache );
	}
	


	/*  Allocate a new SoundInfo object to store information about
	 *  the sound being loaded.
	 */
	try {
		info = new SoundInfo( *cache, name, file );
	}
	catch( audio_error &e ) {
		if( cache->shared == 0 ) delete sounds_by_file.remove(file);
		return ReportException( interp, e );
	}
	
	/*  Add the sound to the name-table so that C++ modules can
	 *  get access to its binary interface.
	 */
	sounds_by_name.put( name, info );
	
	/*  Create a new command to control the sound
	 */
	Tcl_CreateObjCommand( interp, name, SoundCmd, 
						  (ClientData)info, DeleteSound );
	
	/*  Return the name of the new sound.
	 */
	Tcl_SetObjResult( interp, objv[1] );
	return TCL_OK;
}



/*
 *----------------------------------------------------------------------
 *
 * SoundCmd --
 *
 *	This function implements the Tcl commands used to control sound
 *  objects.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Changes the state of the sound object referenced by the ClientData
 *  argument.
 *
 *----------------------------------------------------------------------
 */
static
int SoundCmd( ClientData data, Tcl_Interp *interp,
              int objc, Tcl_Obj *CONST objv[] )
{
	if( objc < 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv, " operation ?param ...?" );
		return TCL_ERROR;
	}
	
	try {
		SoundInfo *info = reinterpret_cast<SoundInfo*>(data);
		int len;
		char *str = Tcl_GetStringFromObj( objv[1], &len );
		
		if( strncmp( "play", str, len ) == 0 ) {
			if( objc > 2 ) {
				Tcl_WrongNumArgs( interp, objc, objv, NULL );
				return TCL_ERROR;
			}

			info->sound->play();

		} else if( strncmp( "loop", str, len ) == 0 ) {
			if( objc > 2 ) {
				Tcl_WrongNumArgs( interp, objc, objv, NULL );
				return TCL_ERROR;
			}

			info->sound->loop();

		} else if( strncmp( "pause", str, len ) == 0 ) {
			if( objc > 2 ) {
				Tcl_WrongNumArgs( interp, objc, objv, NULL );
				return TCL_ERROR;
			}

			info->sound->pause();
			
		} else if( strncmp( "stop", str, len ) == 0 ) {
			if( objc > 2 ) {
				Tcl_WrongNumArgs( interp, objc, objv, NULL );
				return TCL_ERROR;
			}

			info->sound->stop();
			
		} else if( strncmp( "isPlaying", str, len ) == 0 ) {
			if( objc > 2 ) {
				Tcl_WrongNumArgs( interp, objc, objv, NULL );
				return TCL_ERROR;
			}

			Tcl_SetObjResult( interp, Tcl_NewBooleanObj(
			 	static_cast<int>(info->sound->isPlaying()) ) );
			
		} else if( strncmp( "pan", str, len ) == 0 ) {
			if( objc > 3 ) {
				Tcl_WrongNumArgs( interp, objc, objv, "?pan?" );
				return TCL_ERROR;
			}

			Tcl_SetObjResult( interp, Tcl_NewStringObj( 
				"not implemented yet... sorry!", -1 ) );
			return TCL_ERROR;
			
		} else if( strncmp( "frequency", str, len ) == 0 ) {
			if( objc > 3 ) {
				Tcl_WrongNumArgs( interp, objc, objv, "?frequency?" );
				return TCL_ERROR;
			}

			Tcl_SetObjResult( interp, Tcl_NewStringObj( 
				"not implemented yet... sorry!", -1 ) );
			return TCL_ERROR;
			
		} else if( strncmp( "volume", str, len ) == 0 ) {
			if( objc > 3 ) {
				Tcl_WrongNumArgs( interp, objc, objv, "?volume?" );
				return TCL_ERROR;
			}

			Tcl_SetObjResult( interp, Tcl_NewStringObj( 
				"not implemented yet... sorry!", -1 ) );
			return TCL_ERROR;
			
		} else {
			Tcl_Obj *obj = Tcl_NewStringObj( "unknown operation \"", -1 );
			Tcl_AppendToObj( obj, str, len );
			Tcl_AppendToObj( obj, "\"", 1 );
			
			Tcl_SetObjResult( interp, obj );
			return TCL_ERROR;
		}

		return TCL_OK;
	}
	catch( audio_error &e ) {
		return ReportException( interp, e );
	}
}


/*
 *----------------------------------------------------------------------
 *
 * DeleteCmd --
 *
 *	This function implements the "tkwav::destroy" command, used to 
 *  delete Sound objects.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Removes the Sound from the _sounds_by_name and _sounds_by_file
 *  tables, and deletes the Tcl command used to control the Sound.
 *
 *----------------------------------------------------------------------
 */
static 
int DeleteCmd( ClientData data, Tcl_Interp *interp, 
               int objc, Tcl_Obj *CONST objv[] )
{
	for( int i = 1; i < objc; i++ ) {
		int len;
		char *name = Tcl_GetStringFromObj( objv[i], &len );
		
		Tcl_DeleteCommand( interp, name );
	}
	
	return TCL_OK;
}



/*
 *----------------------------------------------------------------------
 *
 * CoopCmd --
 *
 *	This function implements the "tkwav::cooperate" command, used to 
 *  control how the Tk application shares the audio device with other 
 *  applications.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Changes the cooperative level and window-handle of 'device'.
 *
 *----------------------------------------------------------------------
 */
static 
int CoopCmd( ClientData data, Tcl_Interp *interp, 
             int objc, Tcl_Obj *CONST objv[] )
{
	static char *normal = "normal";
	static char *priority = "priority";
	static char *exclusive = "exclusive";
	static char *write_primary = "write_primary";
	

	if( objc > 3 ) {
		Tcl_WrongNumArgs( interp, 1, objv, " ?level? ?window?" );
		return TCL_ERROR;
	} if( objc == 1 ) {
		Tcl_Obj *resultv[2];
		
		switch( coop_level )
		{
		case OutputDevice::NORMAL:
			resultv[0] = Tcl_NewStringObj( normal, -1 );
			break;
		case OutputDevice::PRIORITY:
			resultv[0] = Tcl_NewStringObj( priority, -1 );
			break;
		case OutputDevice::EXCLUSIVE:
			resultv[0] = Tcl_NewStringObj( exclusive, -1 );
			break;
		case OutputDevice::WRITE_PRIMARY:
			resultv[0] = Tcl_NewStringObj( write_primary, -1 );
			break;
		}
		
		resultv[1] = Tcl_NewStringObj( Tk_PathName(coop_window), -1 );

		Tcl_SetObjResult( interp, Tcl_NewListObj(2,resultv) );
		return TCL_OK;

	} else { // objc == 2 || objc == 3
		OutputDevice::CooperativeLevel level;
		Tk_Window win;

		int len;
		char *str = Tcl_GetStringFromObj( objv[1], &len );

		if( strncmp( normal, str, len ) == 0 ) {
			level = OutputDevice::NORMAL;

		} else if( strncmp( priority, str, len ) == 0 ) {
			level = OutputDevice::PRIORITY;

		} else if( strncmp( exclusive, str, len ) == 0 ) {
			level = OutputDevice::EXCLUSIVE;

		} else if( strncmp( write_primary, str, len ) == 0 ) {
			level = OutputDevice::WRITE_PRIMARY;

		} else {
			Tcl_Obj *msg = Tcl_NewStringObj( 
				"unknown cooperative level \"", -1 );
			Tcl_AppendToObj( msg, str, len );
			Tcl_AppendToObj( msg, "\"", -1 );

			Tcl_SetObjResult( interp, msg );
			return TCL_ERROR;
		}

		if( objc == 3 ) {
			str = Tcl_GetStringFromObj( objv[2], &len );

			win = Tk_NameToWindow( interp, str, Tk_MainWindow(interp) );
			if( win == NULL ) return TCL_ERROR;

		} else {
			win = coop_window;
		}

		try {
			device->cooperate( ToplevelWindow(win), level );
			coop_level = level;
			coop_window = win;
			return TCL_OK;
		}
		catch( audio_error &e ) {
			return ReportException( interp, e );
		}
	}
}



/*
 *----------------------------------------------------------------------
 *
 * ReportException --
 *
 *	This function sets the Tcl result of interp to hold the error
 *  message held by the exception e.
 *
 * Results:
 *	Always returns TCL_ERROR, so can be used as:
 *		return ReportException( interp, e );
 *
 * Side effects:
 *	Sets the result of interp.
 *
 *----------------------------------------------------------------------
 */
static int ReportException( Tcl_Interp *interp, std::exception &e ) {
	string msg(e.what());

	Tcl_SetObjResult( interp, 
		Tcl_NewStringObj( const_cast<char*>(msg.c_str()), msg.size() ) );

	return TCL_ERROR;
}


/*----------------------------------------------------------------------
 *
 * DeleteSound --
 *
 *	This function is called back by the Tcl interpreter when the command
 *  controlling a sound is deleted.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Destroys the sound pointed to by the ClientData argument.
 *
 *----------------------------------------------------------------------
 */
static void DeleteSound( ClientData d ) {
	SoundInfo *si = reinterpret_cast<SoundInfo*>(d);
	sounds_by_name.remove( si->name );
	
	FileInfo *fi = sounds_by_file.find( si->file );
	fi->shared--;
	if( fi->shared == 0 ) {
		sounds_by_file.remove( si->file );
		delete fi;
	}
	
	delete si;
}



/*----------------------------------------------------------------------
 *
 * ToplevelWindow --
 *
 *	Returns the Win32 handle of the ancestor of 'win' that is not a
 *  child window.  I.e: the returned window can be used to control
 *  the cooperative level of the audio device.
 *	
 * Results:
 *	The handle (HWND) of the toplevel window containing 'win'.
 *
 * Side effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
static HWND ToplevelWindow( Tk_Window tk_win ) {
	HWND win = Tk_GetHWND( Tk_WindowId(tk_win) );

	while( GetWindowLong( win, GWL_STYLE ) & WS_CHILD ) {
		win = GetParent(win);
	}
	return win;
}

