/*

Tkwav

Copyright (c) 1998, Nat Pryce

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in
all copies of this software.

IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, 
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF 
THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE AUTHOR HAS BEEN ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" 
BASIS, AND THE AUTHOR HAS NO OBLIGATION TO PROVIDE  MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

*/

#include <vector>

#include "wavio.h"
#include "autoclose.h"



using namespace std;



/*  Throw an exception if the MMRESULT indicates an error.
 */
inline 
void Check( MMRESULT status, const char *msg ) throw(audio_error) {
	if( status != MMSYSERR_NOERROR ) throw audio_error( status, msg );
}

struct CloseHMMIO {
	void operator() ( HMMIO h ) { if(h) mmioClose(h,0); }
};



Sound *LoadWavFile( OutputDevice &dev, string file_name )
	throw(audio_error)
{
	AutoClose<HMMIO,CloseHMMIO> file(NULL);
	MMCKINFO riff_chunk;
	LONG bytes;
	
	
	file = mmioOpen( const_cast<char*>(file_name.c_str()), NULL, 
					 MMIO_READ|MMIO_DENYWRITE|MMIO_ALLOCBUF );
	if( !file ) {
		throw audio_error( 0, "failed to open WAV file" );
	}
	
	riff_chunk.fccType = mmioFOURCC( 'W', 'A', 'V', 'E' );
	Check( mmioDescend( file, &riff_chunk, NULL, MMIO_FINDRIFF ),
	       "file not in WAV format" );
	
	
	/*  Find the format of the sound data in the file
	 */
	PCMWAVEFORMAT fmt;

	MMCKINFO fmt_chunk;
	fmt_chunk.ckid = mmioFOURCC( 'f', 'm', 't', ' ' );
	Check( mmioDescend( file, &fmt_chunk, &riff_chunk, MMIO_FINDCHUNK ),
		   "no format information in WAV file" );
	
	/*  Check that the format data is the right size
	 */
	if( fmt_chunk.cksize != sizeof(PCMWAVEFORMAT) ) {
		throw audio_error( 0, "format data is the wrong size" );
	}

	bytes = mmioRead( file, reinterpret_cast<LPSTR>(&fmt), sizeof(fmt) );
	
	Check( mmioAscend( file, &fmt_chunk, 0 ),
		   "failed to end of format information in WAV file" );
	
	
	/*  Find the start of the wave data
	 */
	MMCKINFO data_chunk;
	data_chunk.ckid = mmioFOURCC( 'd', 'a', 't', 'a' );
	Check( mmioDescend( file, &data_chunk, &riff_chunk, MMIO_FINDCHUNK ),
           "failed to find data in WAV file" );
	
	/*  Set up a SoundDescriptor defining the format we want for our
	 *  new sound.
	 */
	SoundDescriptor desc( fmt, data_chunk.cksize );
	
	vector<char> buf(data_chunk.cksize);

	bytes = mmioRead( file, static_cast<LPSTR>(&buf[0]), buf.size() );
	
	if( bytes != buf.size() ) {
		throw audio_error( 0, "failed to read sound data" );
	}

	return new Sound( dev, desc, &buf[0], buf.size() );
}

